#pragma once
#include <J3ML/LinearAlgebra/Vector4.hpp>
#include <J3ML/LinearAlgebra/Vector3.hpp>
#include <J3ML/LinearAlgebra/DirectionVector.hpp>
#include <J3ML/Geometry/Frustum.hpp>
#include <Color4.hpp>

namespace JGL {
    class LightBase;
    class PointLight;
    class SpotLight;
}

// TODO this can be consolidated.
class JGL::LightBase {
protected:
    /* The 4th number represents whether the light is a positional light.
     * In OpenGL, You can have positional lights, Like Point Lights or Spot Lights,
     * Or global lights, Like the sun. */
    Vector4 position = {0, 0, 0, 1};
    Color4 ambient = {0, 0, 0, 0};
    Color4 diffuse = {0, 0, 0, 0};
    Color4 specular = {0, 0, 0, 0};

    float constant_attenuation = 1;
    float linear_attenuation = 0;
    float quadratic_attenuation = 0;
public:
    [[nodiscard]] Vector4 GetPosition() const;
    [[nodiscard]] Color4 GetAmbient() const;
    [[nodiscard]] Color4 GetDiffuse() const;
    [[nodiscard]] Color4 GetSpecular() const;

    [[nodiscard]] float GetConstantAttenuation() const;
    [[nodiscard]] float GetLinearAttenuation() const;
    [[nodiscard]] float GetQuadraticAttenuation() const;
public:
    /// Runs a calculation to determine the lights influence on a given point in 3D space.
    /// @note 0 would be no impact, 1 would be the light is at the same position.
    [[nodiscard]] virtual float GetAttenuationAtPosition(const Vector3& pos) const { return 0; }
public:
    virtual ~LightBase() = default;
};

/// Omni-directional lights.
class JGL::PointLight : public LightBase {
public:
    [[nodiscard]] float GetAttenuationAtPosition(const Vector3& pos) const final;
public:
    PointLight(const Vector3& position, const Color4& ambient, const Color4& diffuse, const Color4& specular, float constant_attenuation = 1, float linear_attenuation = 0, float quadratic_attenuation = 0);
};

/// Lights which only effect things in a given cone.
// TODO get attenuation at position for this.
class JGL::SpotLight : public LightBase {
protected:
    Matrix3x3 orientation;
    float exponent;
    float cut;
public:
    /// Create a spotlight in 3D space.
    /// @param position The position of the light in 3D space.
    /// @param ro_mat Orientation of the light in 3D space.
    /// @param cone_size_degrees The size of the cone.
    /// @param exponent How focused the beam should be, Higher is more focused, Lower is less.
    /// @param ambient How much this light should effect the ambient light of the scene.
    /// @param diffuse
    /// @param specular How much this light should effect specular highlights of objects being influenced by it.
    SpotLight(const Vector3& position, const Matrix3x3& ro_mat, float cone_size_degrees, float exponent, const Color4& ambient, const Color4& diffuse, const Color4& specular, float constant_attenuation = 1, float linear_attenuation = 0, float quadratic_attenuation = 0);
};