#pragma once
#include <vector>
#include <glad/glad.h>
#include <Color4.hpp>
#include <Colors.hpp>
#include <JGL/types/Enums.h>
#include <J3ML/LinearAlgebra/Vector2i.hpp>

namespace JGL {
    class RenderTarget;
    class Texture;
}

using J3ML::LinearAlgebra::Vector2i;
class JGL::RenderTarget {
private:
    Color4 clear_color{0,0,0,0};
    Vector2i size{0, 0};
    bool using_depth = false;
    bool texture_created_by_us = false;
    GLuint framebuffer_object = 0;
    GLuint depth_buffer = 0;
    const Texture* texture = nullptr;
    MSAA_SAMPLE_RATE msaa_sample_rate = MSAA_SAMPLE_RATE::MSAA_NONE;
    GLuint msaa_framebuffer_object = 0;
    GLuint msaa_depth_buffer = 0;
    GLuint msaa_render_buffer = 0;
    void Erase();
public:
    /// @returns The Render Target currently in use by OpenGL.
    /// @note Zero is the screen.
    static GLuint GetActiveGLFramebufferHandle();

    /// Changes the Render Target that OpenGL will draw on.
    /// @param render_target The new Render Target for OpenGL to draw on.
    static void SetActiveGLRenderTarget(const RenderTarget& render_target);

    /// Changes the size of the renderable area of this Render Target.
    /// @param new_size new width & height in pixels.
    /// @note The data stored in this Render Target will be lost.
    void Resize(const Vector2i& new_size);

    /// Sets the MSAA mode for this Render Target.
    /// @returns false if the mode isn't available, true for success.
    [[nodiscard]] bool SetMSAAEnabled(MSAA_SAMPLE_RATE sample_rate);

    /// If you're using MSAA and not using J2D || J3D Begin & End you must do this.
    void MSAABlit() const;

    void RegenerateMipMaps();
    /// Copy one Render Target onto another. Will break if they're not the same size.
    // TODO support different sizes. If the destination is too small fix it for them but log a warning.
    static void Blit(const RenderTarget& source, RenderTarget* destination, const Vector2i& position = {0, 0});

    /// Plots a single pixel onto a Render Target.
    /// @param color The color to render.
    /// @param position The position in the destination to draw the pixel.
    /// @param destination The destination RenderTarget.
    static void Blit(const Color4& color, const Vector2i& position, RenderTarget* destination);

    /// Blit an input texture onto this render target at the given position.
    /// @param source Source texture.
    /// @param destination Render Target to draw on.
    /// @param position Where in the destination to draw.
    static void Blit(const Texture* source, RenderTarget* destination, const Vector2i& position = {0, 0});

    /// @returns Whether or not this Render Target created it's Texture.
    [[nodiscard]] bool OwnsTexture() const;
public:
    /// @returns the size of the renderable area.
    [[nodiscard]] Vector2i GetDimensions() const;

    /// @returns The currently selected MSAA Sample Rate.
    [[nodiscard]] MSAA_SAMPLE_RATE GetMSAASampleRate() const;

    /// @returns Whether or not this Render Target is using MSAA.
    [[nodiscard]] bool MSAAEnabled() const;

    /// @returns The JGL texture this Render Target draws on.
    [[nodiscard]] const Texture* GetTexture() const;

    /// @returns The OpenGL handle for the texture this Render Target draws on.
    [[nodiscard]] GLuint GetTextureHandle() const;

    /// @returns The OpenGL handle for this Render Target.
    [[nodiscard]] GLuint GetGLFramebufferObjectHandle() const;

    /// @returns The handle to the OpenGL buffer containing depth information
    /// @note Only valid if this Render Target is being used for 3D.
    [[nodiscard]] GLuint GetGLDepthBufferHandle() const;

    /// @returns The color that should be used to clear this Render Target.
    [[nodiscard]] Color4 GetClearColor() const;

    /// @returns The color information for this Render Target.
    /// @note The CPU thread this is called from & the GPU cannot do anything while this takes place. It's very slow.
    [[nodiscard]] std::vector<GLfloat> GetPixels() const;

    [[nodiscard]] static Vector2i MaximumSize();
public:
    /// Create a Render Target from a Render Target that already exists.
    /** @note Render Targets that are copies of another will copy the Texture.
     * This is so that deleting the copy doesn't delete the Texture of the original.
    */
    RenderTarget(const RenderTarget& rhs);

    /// Create a Render Target for a texture that already exists.
    /// @param texture The Texture that using this Render Target would draw on.
    /// @param clear_color The color to be used if you want to clear the Render Target.
    /// @note For Render Targets created this way, The destructor will not delete the texture.
    explicit RenderTarget(const Texture* texture, const Color4& clear_color = Colors::Transparent);

    /// Create a Render Target with a new texture.
    /// @param size The width & height the Render Target should have.
    /// @param clear_color The color to be used if you want to clear the Render Target.
    /// @param use_depth Whether or not this Render Target will have depth information.
    /// @param sample_rate The MSAA sample rate this Render Target will use.
    explicit RenderTarget(const Vector2i& size, const Color4& clear_color = Colors::Transparent, bool use_depth = false,
                          MSAA_SAMPLE_RATE sample_rate = MSAA_SAMPLE_RATE::MSAA_NONE, FilteringMode filteirng_mode = FilteringMode::NEAREST);

    /// Deletes this Render Target.
    /** @note If this Render Target was made with a Texture that already existed
     * the Texture will not be deleted. */
    ~RenderTarget();

};