#pragma once
#include<vector>
#include <filesystem>
#include <J3ML/LinearAlgebra/Vector2i.hpp>
#include <Color3.hpp>
#include <Color4.hpp>
#include <JGL/types/Enums.h>

namespace JGL {
    using J3ML::LinearAlgebra::Vector2i;
    class Texture;
}

/// Represents texture data loaded on the GPU. Contains a handle that can be passed to OpenGL draw calls.
class JGL::Texture {
protected:
    unsigned int texture_handle = 0;
    bool invert_y = false;
    Vector2i size = {0, 0};
    ColorFormat format =  ColorFormat::RGBA;
    FilteringMode filtering_mode;
    WrappingMode wrapping_mode;
    void load(const unsigned char* pixels);
    std::vector<unsigned char> png(const std::filesystem::path& file);
    std::vector<unsigned char> bmp(const std::filesystem::path& file);
    [[nodiscard]] bool SizeExceedsMaximum(const Vector2i& size);
public:
    /// @returns A handle used to identify this texture.
    [[nodiscard]] unsigned int GetHandle() const;
    /// @returns The size of this texture.
    [[nodiscard]] Vector2i GetDimensions() const;
    /// @returns The filtering mode this texture is using.
    [[nodiscard]] FilteringMode GetFilteringMode() const;
    /// @returns The way this texture behaves when used on geometry of different sizes.
    [[nodiscard]] WrappingMode GetWrappingMode() const;
    /// @returns The orientation of this texture in v-ram.
    /// @note true is right-side-up because OpenGL defaults to upside-down.
    [[nodiscard]] bool Inverted() const;
    [[nodiscard]] ColorFormat GetFormat() const;
    /// @returns The raw pixels this texture is made up of.
    /// @note This will read-back from the GPU. Slow.
    [[nodiscard]] std::vector<Color4> GetPixelData() const;
    /// @returns The biggest size for a texture on this system.
    /// @note on modern systems this is *usually* ridiculous.
    [[nodiscard]] static Vector2i MaximumSize();
public:
    /// Load a texture from a file,
    explicit Texture(const std::filesystem::path& file, FilteringMode filtering_mode = FilteringMode::BILINEAR, WrappingMode wrapping_mode = WrappingMode::CLAMP_TO_EDGE, bool invert_y = true);
    /// Load a texture from raw pixels.
    Texture(const Color4* pixels, const Vector2i& size, const ColorFormat& format, FilteringMode filtering_mode, WrappingMode wrapping_mode);
    /// Load a texture from raw pixels.
    Texture(const Color3* pixels, const Vector2i& size, const ColorFormat& format, FilteringMode filtering_mode, WrappingMode wrapping_mode);
    /// Construct a Texture Atlas from many different textures.
    /// @note THIS IS UNFINISHED.
    Texture(const Texture* textures, const size_t& texture_count);
    /// Initialize a texture filled with trash data.
    /// @see RenderTarget
    explicit Texture(const Vector2i& size, FilteringMode filtering_mode = FilteringMode::NEAREST);

    Texture(const Texture& rhs);
    ~Texture();
};