#pragma once

#include <glad/glad.h>
#include <array>
#include <vector>
#include <J3ML/LinearAlgebra/Vector2.hpp>
#include <JGL/types/RenderTarget.h>
#include <JGL/types/Light.h>
#include <JGL/logger/logger.h>

namespace JGL {
    class State;
    class StateStack;
}

class JGL::State {
public:
    GLfloat clear_color[4] = {0, 0, 0, 1};
    GLfloat draw_color[4] = {1, 1, 1, 1};
    GLint viewport[4] = {0, 0, 0, 0};
    GLint blend_func[2];
    GLuint current_fbo = 0;
    RenderTarget* current_render_target = nullptr;

    bool texture_2D = false;
    bool texture_coordinate_array = false;
    bool normal_array = false;
    bool depth_test = false;
    bool vertex_array = false;
    bool cull_face = false;
    bool blend = false;
    bool color_array = false;
    GLint selected_texture_unit = 0;
    // List of lights required for each object in the scene. up-to 8. For example, the sun. Or a flash-light.
    std::array<const LightBase*, 8> required_lights{};
    // List of all lights in the scene.
    std::vector<const LightBase*> optional_lights{};
public:
    static State SaveState();
    static void RestoreState(const State& state);
};

class JGL::StateStack {
private:
    std::vector<State> states{};
public:
    size_t Size() { return states.size(); }
    void Push(const State& state);
    void Pop();
    State* PreviousState();
public:
    StateStack() = default;
    ~StateStack() = default;
};

namespace JGL {
    inline constexpr std::array<const LightBase*, 8> empty_light_array = { nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr };
    std::vector<Vector3> TriangleMeshVertexNormals(const Vector3* vertices, const size_t& vertex_count, const unsigned int* indices, const size_t& index_count);
    inline StateStack state_stack;
    inline Vector2i window_size;
}

namespace JGL::J2D {
    inline State default_state
    {
            {0, 0, 0, 1},
            {1, 1, 1, 1},
            {0, 0, 0, 0},
            {GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA},
            0, nullptr, false, false,
            false, false, true, true,
            true, false, 0,
            {}, {}
    };

    inline State current_state;
}

namespace JGL::J3D {
    inline State current_state;
    inline float far_plane = 0;
    inline float fov = 0;
    // Enables lighting and selects the correct lights to use.
    void SelectLights(const Vector3& position);
    // Resets the GL lights to default and disables them. Then, disables lighting.
    void ResetLights();
    [[nodiscard]] bool UsingLighting();
}